/**
 * Copyright 2006-8 Simon Andrews
 *
 *    This file is part of ChIPMonk.
 *
 *    ChIPMonk is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    ChIPMonk is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with ChIPMonk; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package uk.ac.bbsrc.babraham.ChipMonk.Displays.ChromosomeViewer;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Image;
import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import java.awt.image.BufferedImage;
import java.util.Enumeration;
import java.util.Vector;

import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollBar;
import javax.swing.JScrollPane;

import uk.ac.bbsrc.babraham.ChipMonk.ChipMonkApplication;
import uk.ac.bbsrc.babraham.ChipMonk.DataTypes.Chromosome;
import uk.ac.bbsrc.babraham.ChipMonk.DataTypes.DataStore;

public class ChromosomeViewer extends JPanel{

	private ChipMonkApplication application;
	private Chromosome chromosome;
	private double basesPerPixel = 10000;
	private Vector featureTracks = new Vector();
	private Vector dataTracks = new Vector();
	private RefreshScrollPane scrollPane;
	private FeaturePanel featurePanel;
	private JLabel titleLabel;
	private int selectionStart = 0;
	private int selectionEnd = 0;
	private boolean makingSelection = false;
	private static boolean showAllLables = false;
	private static double maxLogRatio = 1;
	private int currentStart = 1;
	private int currentEnd = 1;

	public ChromosomeViewer (ChipMonkApplication application, Chromosome chromosome) {
		this.application = application;
		this.chromosome = chromosome;
		
		setLayout(new BorderLayout());
		titleLabel = new JLabel(application.genome().species()+" "+application.genome().assembly()+" Chromosome "+chromosome.name()+" "+chromosome.length()+"bp",JLabel.CENTER); 
		add(titleLabel,BorderLayout.NORTH);
		
		featurePanel = new FeaturePanel();
		featurePanel.setLayout(new GridBagLayout());
				
		scrollPane = new RefreshScrollPane(featurePanel,JScrollPane.VERTICAL_SCROLLBAR_NEVER,JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
		add(scrollPane,BorderLayout.CENTER);
		tracksUpdated();
	}
		
	public void tracksUpdated() {
		
		String [] featureTypes = application.drawnFeatureTypes();
		featureTracks.removeAllElements();
		for (int i=0;i<featureTypes.length;i++) {
			ChromosomeFeatureTrack t = new ChromosomeFeatureTrack(this,featureTypes[i]);
			featureTracks.add(t);
		}
		
		DataStore [] dataStores = application.drawnDataStores();
		dataTracks.removeAllElements();
		for (int i=0;i<dataStores.length;i++) {
			ChromosomeDataTrack t = new ChromosomeDataTrack(this,application.dataCollection(),dataStores[i]);
			dataTracks.add(t);
		}

		
		featurePanel.removeAll();
		GridBagConstraints c = new GridBagConstraints();
		c.gridx=0;
		c.gridy=0;
		c.fill = GridBagConstraints.BOTH;
		c.weightx = 1;
		c.weighty = 0.2;

		Enumeration e = featureTracks.elements();
		while (e.hasMoreElements()) {
			featurePanel.add((ChromosomeFeatureTrack)e.nextElement(),c);
			c.gridy++;
		}
		
		// We weight the data tracks twice as heavily as the feature tracks
		c.weighty = 0.4;
		e = dataTracks.elements();
		while (e.hasMoreElements()) {
			featurePanel.add((ChromosomeDataTrack)e.nextElement(),c);
			c.gridy++;
		}
		featurePanel.validate();
		scrollPane.revalidate();
		featurePanel.repaint();
	}
	
	public void setView (int start, int end) {
		
		// If the view is a reversed section we need to swap
		// start and end
		if (start > end) {
			int temp = start;
			start = end;
			end = temp;
		}
		
		int width = scrollPane.getVisibleRect().width;
		basesPerPixel = (double)(end-start)/width;
		remove(scrollPane);
		scrollPane = new RefreshScrollPane(featurePanel,JScrollPane.VERTICAL_SCROLLBAR_NEVER,JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
		add(scrollPane,BorderLayout.CENTER);
		validate();
		setMidPoint (start + ((end-start)/2));
	}
	
	public void setMidPoint (int bp) {
		JScrollBar b = scrollPane.getHorizontalScrollBar();
		int width = scrollPane.getVisibleRect().width;
		int newValue = (int)(((double)bp)/basesPerPixel)-width/2;
		if (newValue < b.getMinimum()) {
			newValue = b.getMinimum();
		}
		if (newValue > b.getMaximum()) {
			newValue = b.getMaximum();
		}
		b.setValue(newValue);
		
		
	}
	
	public void toggleLabels () {
		if (showAllLables) {
			showAllLables = false;
		}
		else {
			showAllLables = true;
		}
		tracksUpdated();
	}
	
	
	public boolean showAllLables () {
		return showAllLables;
	}
	
	public void zoomOut () {
		int midBase = getMidPoint();
		basesPerPixel *= 2;
		// The actual limit on zoom is provided by the feature panel
		// in the paintComponent method which updates the genome view
		// and the top status bar.  This check is just to stop the
		// basesPerPixel value getting bigger than an int can hold.
		if (basesPerPixel > chromosome.length())
			basesPerPixel = chromosome.length();
		
		remove(scrollPane);
		scrollPane = new RefreshScrollPane(featurePanel,JScrollPane.VERTICAL_SCROLLBAR_NEVER,JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
		add(scrollPane,BorderLayout.CENTER);
		validate();
		setMidPoint(midBase);
	}

	public void zoomIn () {
		int midBase = getMidPoint();
		basesPerPixel /= 2;
		// The actual limit on zoom is provided by the feature panel
		// in the paintComponent method which updates the genome view
		// and the top status bar.  This check is just to stop the
		// basesPerPixel value getting bigger than an int can hold.
		if (basesPerPixel < 0.1)
			basesPerPixel = 0.1;
		
		remove(scrollPane);
		scrollPane = new RefreshScrollPane(featurePanel,JScrollPane.VERTICAL_SCROLLBAR_NEVER,JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
		add(scrollPane,BorderLayout.CENTER);
		validate();
		setMidPoint(midBase);
	}

	public void moveLeft() {
		scrollPane.getHorizontalScrollBar().setValue(scrollPane.getHorizontalScrollBar().getValue()-scrollPane.getHorizontalScrollBar().getBlockIncrement());
	}
	public void moveRight() {
		scrollPane.getHorizontalScrollBar().setValue(scrollPane.getHorizontalScrollBar().getValue()+scrollPane.getHorizontalScrollBar().getBlockIncrement());
	}
	
	public int getMidPoint() {
		int midX = 	(int)(featurePanel.getVisibleRect().getX() + (featurePanel.getVisibleRect().width/2));
		return (int)(midX * basesPerPixel);
	}
	
	public void setMakingSelection (boolean b) {
		makingSelection = b;
		repaint();
	}
	
	public void setSelectionStart (int x) {
		selectionStart = x;
		repaint();
	}
	
	public void setSelectionEnd (int x) {
		selectionEnd = x;
		repaint();
	}
	
	public boolean makingSelection () {
		return makingSelection;
	}
	
	public int selectionStart () {
		return selectionStart;
	}
	
	public int selectionEnd () {
		return selectionEnd;
	}
	
	public int currentStart () {
		return currentStart;
	}
	
	public int currentEnd () {
		return currentEnd;
	}
	
	public double maxLogRatio () {
		return maxLogRatio;
	}

	public double minLogRatio () {
		return 0-maxLogRatio;
	}
	
	public void setMaxLogRatio (double d) {
		maxLogRatio = d;
		tracksUpdated();
	}

	public ChipMonkApplication application () {
		return application;
	}
	
	public int getIndex (ChromosomeFeatureTrack t) {
		return featureTracks.indexOf(t);
	}

	public int getIndex (ChromosomeDataTrack t) {
		return dataTracks.indexOf(t);
	}

	
	public Chromosome chromosome () {
		return chromosome;
	}
	
	public void setBasesPerPixel (double l) {
			basesPerPixel = l;
			scrollPane.revalidate();
			repaint();
	}
		
	private class FeaturePanel extends JPanel  {
		
		public Dimension getPreferredSize () {
			int width = (int)((double)chromosome.length()/basesPerPixel);
			int height = getVisibleRect().height;
			
//			System.out.println("Requesting size "+width+"x"+height);
			return new Dimension(width,height);
			
		}
		public void paintComponent (Graphics g) {
			super.paintComponent(g);
			// Since we're repainting we should update the
			// information about what we're looking at
			int x = getVisibleRect().x;
			int length = getVisibleRect().width;
			
//			System.out.println("BPP:"+basesPerPixel+" X:"+x+" length:"+length);
			
			currentStart = (int)(basesPerPixel*x);
			currentEnd = (int)(basesPerPixel*(x+length));
			if (currentEnd > chromosome.length())
				currentEnd = chromosome.length();
			titleLabel.setText(application.genome().species()+" "+application.genome().assembly()+" Chromosome "+chromosome.name()+" "+currentStart+"-"+currentEnd+"bp"+" ("+(currentEnd-currentStart)+"bp)");
			// Update the genome view so the red box moves with us
			application.setChromosomeView(chromosome,currentStart,currentEnd);			
		}
	}
	
	public Image getCurrentView() {
		BufferedImage b = new BufferedImage(getWidth(),getHeight(),BufferedImage.TYPE_INT_RGB);
		Graphics g = b.getGraphics();
		
		// We don't want the scroll bars to appear in the drawn image
		// so we temporarily remove them.
		scrollPane.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
		scrollPane.validate();
		this.paint(g);
		
		// Remember to put the scroll bars back after we've done the drawing.
		scrollPane.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
		scrollPane.validate();
		
		return b;
	}
	
	private class RefreshScrollPane extends JScrollPane {
		
		// This class is needed in order to keep the static labels on the
		// feature and data tracks in the same place whilst scrolling.  If
		// I don't add this extra listener then you get wierd ghosting
		// effects when scrolling and the static labels move with the track.
		//
		// There's probably a more efficient way to do this than this, but
		// for the time being it works and is better than the alternative!
		
		public RefreshScrollPane (Component c, int h, int v) {
			super (c,h,v);
			getHorizontalScrollBar().setBlockIncrement(10);
			JScrollBar b = getHorizontalScrollBar();
			b.addAdjustmentListener(new AdjustmentListener() {
				public void adjustmentValueChanged(AdjustmentEvent e) {
					getViewport().repaint();
				}
			});
		}
	}
}
